#include <SDL.h>
#include <SDL_image.h>
#include "def.h"
#include "log.h"
#include "opengl.h"

// Recommended for compatibility bullshit.
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
#  define RMASK 0xff000000
#  define GMASK 0x00ff0000
#  define BMASK 0x0000ff00
#  define AMASK 0x000000ff
#else
#  define RMASK 0x000000ff
#  define GMASK 0x0000ff00
#  define BMASK 0x00ff0000
#  define AMASK 0xff000000
#endif
#define RGBMASK RMASK,GMASK,BMASK,AMASK

// The screen info, gives data of current opengl settings.
gl_info gl_screen;

// Our precious camera.
Vec2* gl_camera;

static int flip_surface(SDL_Surface* surface);

// Flips the surface vertically. Return 0 on success.
static int flip_surface(SDL_Surface* surface) {
  // Flip the image.
  Uint8* rowhi, *rowlo, *tmpbuf;
  int y;

  tmpbuf = (Uint8*)malloc(surface->pitch);
  if(tmpbuf == NULL) {
    WARN("Out of memory");
    return -1;
  }

  rowhi = (Uint8*)surface->pixels;
  rowlo = rowhi + (surface->h * surface->pitch) - surface->pitch;
  for(y = 0; y < surface->h / 2; ++y) {
    memcpy(tmpbuf, rowhi, surface->pitch);
    memcpy(rowhi, rowlo, surface->pitch);
    memcpy(rowlo, tmpbuf, surface->pitch);
    rowhi += surface->pitch;
    rowlo -= surface->pitch;
  }
  free(tmpbuf);
  // Might be obvious, but I'm done flipping.
  return 0;
}

// Load the image directly as an opengl texture.
gl_texture* gl_newImage(const char* path) {
  SDL_Surface* tmp, *surface;
  Uint32 saved_flags;
  Uint8  saved_alpha;
  int potw, poth;

  tmp = IMG_Load(path); // Load the surface.
  if(tmp == 0) {
    WARN("'%s' could not be opened: %s", path, IMG_GetError());
    return 0;
  }

  surface = SDL_DisplayFormatAlpha(tmp); // Set the surface to what we use.
  if(surface == 0) {
    WARN("Error converting image to screen format: %s", SDL_GetError());
    return 0;
  }

  SDL_FreeSurface(tmp); // Free the temp surface.

  flip_surface(surface);

  // Set up the texture defaults.
  gl_texture* texture = MALLOC_L(gl_texture);
  texture->w = (FP)surface->w;
  texture->h = (FP)surface->h;
  texture->sx = 1.0;
  texture->sy = 1.0;

  // Ensure size is power of two.
  potw = surface->w;
  if((potw &(potw-1)) != 0) {
    potw = 1;
    while(potw < surface->w)
      potw <<= 1;
  }
  texture->rw = potw;
  
  poth = surface->h;
  if((poth &(poth-1)) != 0) {
    poth = 1;
    while(poth < surface->h)
      poth <<= 1;
  }
  texture->rh = poth;

  if(surface->w != potw || surface->h != poth) {
    // Size isn't original.
    SDL_Rect rtemp;
    rtemp.x = rtemp.y = 0;
    rtemp.w = surface->w;
    rtemp.h = surface->h;

    // Save alpha.
    saved_flags = surface->flags & (SDL_SRCALPHA | SDL_RLEACCELOK);
    saved_alpha = surface->format->alpha;
    if((saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA)
      SDL_SetAlpha(surface, 0, 0);

    // Create the temp POT surface.
    tmp = SDL_CreateRGBSurface(SDL_SRCCOLORKEY,
          texture->rw, texture->rh, surface->format->BytesPerPixel*8, RGBMASK);
    if(tmp == NULL) {
      WARN("Unable to create POT surface %s", SDL_GetError());
      return 0;
    }
    if(SDL_FillRect(tmp, NULL, SDL_MapRGBA(surface->format, 0, 0, 0, SDL_ALPHA_TRANSPARENT))) {
      WARN("Unable to fill rect: %s", SDL_GetError());
      return 0;
    }

    SDL_BlitSurface(surface, &rtemp, tmp, &rtemp);
    SDL_FreeSurface(surface);

    surface = tmp;

    // Set saved alpha.
    if((saved_flags & SDL_SRCALPHA) == SDL_SRCALPHA)
      SDL_SetAlpha(surface, saved_flags, saved_alpha);
  }

  glGenTextures(1, &texture->texture); // Create the texure.
  glBindTexture(GL_TEXTURE_2D, texture->texture); // Load the texture.

  // Filtering, LINEAR is better for scaling, nearest looks nicer, LINEAR
  // also seems to create a bit of artifacts around the edges.
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  //glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

  SDL_LockSurface(surface);
  glTexImage2D(GL_TEXTURE_2D, 0, surface->format->BytesPerPixel,
        surface->w, surface->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, surface->pixels);

  SDL_UnlockSurface(surface);
  SDL_FreeSurface(surface);

  return texture;
}

// Load the texture immediately, but also set is as a sprite.
gl_texture* gl_newSprite(const char* path, const int sx, const int sy) {
  gl_texture* texture;
  if((texture = gl_newImage(path)) == NULL)
    return NULL;
  texture->sx = (FP)sx;
  texture->sy = (FP)sy;
  texture->sw = texture->w/texture->sx;
  texture->sh = texture->h/texture->sy;
  return texture;
}

// Free the texture.
void gl_free(gl_texture* texture) {
  glDeleteTextures(1, &texture->texture);
  free(texture);
}

// Blit the sprite at given position.
void gl_blitSprite(gl_texture* sprite, Vec2* pos, const int sx, const int sy) {
  glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  glTranslatef(sprite->sw * (FP)(sx)/sprite->rw,
        sprite->sh*(sprite->sy-(FP)sy-1)/sprite->rh, 0.0f);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix(); // Projection translation matrix.
  glTranslatef(gl_camera->x - pos->x - sprite->sw / 2.0,
        gl_camera->y - pos->y - sprite->sh/2.0, 0.0f);

  // Actual blitting....
  glBindTexture(GL_TEXTURE_2D, sprite->texture);
  glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0f, 0.0f);
      glVertex2f(0.0f, 0.0f);
    glTexCoord2f(sprite->sw/sprite->rw, 0.0f);
      glVertex2f(sprite->sw, 0.0f);
    glTexCoord2f(0.0f, sprite->sh/sprite->rh);
      glVertex2f(0.0f, sprite->sh);
    glTexCoord2f(sprite->sw/sprite->rw, sprite->sh/sprite->rh);
      glVertex2f(sprite->sw, sprite->sh);
  glEnd();

  glPopMatrix(); // Projection translation matrix.

  glMatrixMode(GL_TEXTURE);
  glPopMatrix(); // Sprite translation matrix.
}

// Just straight out blit the thing at position.
void gl_blit(gl_texture* texture, Vec2* pos) {
  glMatrixMode(GL_PROJECTION);
  glPushMatrix(); // Set up translation matrix.
  glTranslatef(pos->x, pos->y, 0);
  
  // Actual blitting..
  glBindTexture(GL_TEXTURE_2D, texture->texture);
  glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0f, 0.0f);
      glVertex2f(0.0f, 0.0f);
    glTexCoord2f(texture->w/texture->rw, 0.0f);
      glVertex2f(texture->w, 0.0f);
    glTexCoord2f(0.0f, texture->h/texture->rh);
      glVertex2f(0.0f, texture->h);
    glTexCoord2f(texture->w/texture->rw, texture->h/texture->rh);
      glVertex2f(texture->w, texture->h);
  glEnd();

  glPopMatrix(); // Pop the translation matrix.
}

// Bind our precious camera to a vector.
void gl_bindCamera(Vec2* pos) {
  gl_camera = pos;
}

// Initialize SDL/OpenGL etc.
int gl_init(void) {
  int depth;
  int flags = SDL_OPENGL;

  // Initializes video.
  if(SDL_Init(SDL_INIT_VIDEO) < 0) {
    WARN("Unable to initialize SDL: %s", SDL_GetError());
    return -1;
  }

  // FFUUUU Ugly cursor thing.
  SDL_ShowCursor(SDL_DISABLE);

  flags |= SDL_FULLSCREEN* gl_screen.fullscreen;
  depth = SDL_VideoModeOK(gl_screen.w, gl_screen.h, gl_screen.depth, flags); // Test set up.
  if(depth != gl_screen.depth)
    WARN("Depth: %d bpp unavailable, will use %d bpp", gl_screen.depth, depth);

  gl_screen.depth = depth;

  // Actually creating the screen.
  if(SDL_SetVideoMode(gl_screen.w, gl_screen.h, gl_screen.depth, flags) == NULL) {
    WARN("Unable to create OpenGL window: %s", SDL_GetError());
    SDL_Quit();
    return -1;
  }

  // Grab some info.
  SDL_GL_GetAttribute(SDL_GL_RED_SIZE,      &gl_screen.r);
  SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE,    &gl_screen.g);
  SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE,     &gl_screen.b);
  SDL_GL_GetAttribute(SDL_GL_ALPHA_SIZE,    &gl_screen.a);
  SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER,  &gl_screen.doublebuf);
  gl_screen.depth = gl_screen.r + gl_screen.g + gl_screen.b + gl_screen.a;

  // Debug heaven.
  DEBUG("OpenGL Window Created: %dx%d@%dbpp %s", gl_screen.w, gl_screen.h, gl_screen.depth,
        gl_screen.fullscreen ? "fullscreen" : "window");
  DEBUG("r: %d, g: %d, b: %d, a: %d, doublebuffer: %s", gl_screen.r, gl_screen.g, gl_screen.b, gl_screen.a,
        gl_screen.doublebuf ? "yes" : "no");
  DEBUG("Renderer: %s", glGetString(GL_RENDERER));

  // Some openGL options.
  glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
  glDisable(GL_DEPTH_TEST); // Set for doing 2D shidazles.
  glEnable(GL_TEXTURE_2D);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(-gl_screen.w/2,   // Left edge.
          gl_screen.w/2,    // Right edge.
          -gl_screen.h/2,   // Bottom edge.
          gl_screen.h/2,    // Top edge.
          -1.0f,            // Near.
          1.0f);            // Far.
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // Alpha.
  glEnable(GL_BLEND);

  glClear(GL_COLOR_BUFFER_BIT);

  SDL_WM_SetCaption(WINDOW_CAPTION, NULL);

  return 0;
}

// Clean up our mess.
void gl_exit(void) {
  SDL_ShowCursor(SDL_ENABLE);
  SDL_Quit();
}

